/*
 * The MIT License (MIT)
 *
 * Copyright © 2023 xrv <xrv@live.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.tzld.crawler.etl.mq;

import com.alibaba.fastjson.JSONObject;
import com.aliyun.mq.http.MQClient;
import com.aliyun.mq.http.MQConsumer;
import com.aliyun.mq.http.model.Message;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.tzld.crawler.etl.model.vo.CrawlerVideoVO;
import com.tzld.crawler.etl.service.EtlService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * @author ehlxr
 * @since 2023-06-09 15:24.
 */
@Component
public class EtlMQConsumer {
    private static final Logger log = LoggerFactory.getLogger(EtlMQConsumer.class);
    private final EtlService etlService;
    @Value("${rocketmq.accessKey:}")
    private String accessKey;
    @Value("${rocketmq.secretKey:}")
    private String secretKey;
    @Value("${rocketmq.httpEndpoint:}")
    private String httpEndpoint;
    @Value("${rocketmq.instanceId:}")
    private String instanceId;
    @Value("${rocketmq.crawler.etl.topic:}")
    private String topic;
    @Value("${rocketmq.crawler.etl.groupid:}")
    private String groupId;

    @ApolloJsonValue("${crawler.etl.priority:{}}")
    private Map<String, Integer> priorityMap;

    private MQClient mqClient;
    private ThreadPoolExecutor priorityPool;

    public EtlMQConsumer(EtlService etlService) {
        this.etlService = etlService;
    }

    @PostConstruct
    public void init() {
        mqClient = new MQClient(httpEndpoint, accessKey, secretKey);
        priorityPool = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors(), Runtime.getRuntime().availableProcessors() * 2, 60L,
                TimeUnit.SECONDS, new PriorityBlockingQueue<>(),
                new ThreadFactoryBuilder().setNameFormat("priority-etl-pool-%d").build());

        new Thread(this::consumeMsg).start();
    }

    class RunnablePriority implements Runnable, Comparable<RunnablePriority> {
        private final CrawlerVideoVO video;

        public RunnablePriority(CrawlerVideoVO video) {
            this.video = video;
        }

        public Integer getPriority() {
            return priorityMap.getOrDefault(video.getPlatform(), -1);
        }

        /**
         * 优先级比较
         */
        @Override
        public int compareTo(RunnablePriority o) {
            if (this.getPriority() < o.getPriority()) {
                return 1;
            } else if (this.getPriority() > o.getPriority()) {
                return -1;
            }
            return 0;
        }

        @Override
        public void run() {
            try {
                log.info("Thread pool info \nTaskCount: {}\nCompletedTaskCount: {}\nActiveCount: {}\nQueue size: {}\nQueue elements: {}",
                        priorityPool.getTaskCount(), priorityPool.getCompletedTaskCount(), priorityPool.getActiveCount(),
                        priorityPool.getQueue().size(), priorityPool.getQueue());
                log.info("deal video: {} priority: {}", video, getPriority());
                etlService.deal(video);
                log.info("Thread pool info \nTaskCount: {}\nCompletedTaskCount: {}\nActiveCount: {}\nQueue size: {}\nQueue elements: {}",
                        priorityPool.getTaskCount(), priorityPool.getCompletedTaskCount(), priorityPool.getActiveCount(),
                        priorityPool.getQueue().size(), priorityPool.getQueue());
            } catch (Exception e) {
                log.error("deal video {} error.", video, e);
            }
        }

        @Override
        public String toString() {
            return video.getPlatform() + ":" + video.getOutVideoId();
        }
    }

    private void consumeMsg() {
        final MQConsumer consumer = mqClient.getConsumer(instanceId, topic, groupId, null);
        do {
            List<String> handles = new ArrayList<>();
            try {
                List<Message> messages = consumer.consumeMessage(10, 10);
                if (messages == null || messages.isEmpty()) {
                    log.info("No new message, continue");
                    continue;
                }

                messages.forEach(message -> {
                    log.info("Receive message: {} from topic: {}, group: {}", message, topic, groupId);
                    handles.add(message.getReceiptHandle());
                    CrawlerVideoVO video = JSONObject.parseObject(message.getMessageBodyString(), CrawlerVideoVO.class);
                    priorityPool.execute(new RunnablePriority(video));
                });
            } catch (Throwable e) {
                log.error("Consume message from topic: {}, group: {} error", topic, groupId, e);
            }
            if (!handles.isEmpty()) {
                consumer.ackMessage(handles);
            }
        } while (true);
    }
}
