/*
 * The MIT License (MIT)
 *
 * Copyright © 2023 xrv <xrv@live.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.tzld.crawler.etl.service.impl;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import com.tzld.commons.aliyun.oss.AliyunOssManager;
import com.tzld.crawler.etl.common.base.CommonResponse;
import com.tzld.crawler.etl.common.base.Constant;
import com.tzld.crawler.etl.common.enums.ExceptionEnum;
import com.tzld.crawler.etl.common.exception.CommonException;
import com.tzld.crawler.etl.dao.mapper.CrawlerVideoMapper;
import com.tzld.crawler.etl.model.dto.StrategyDataDto;
import com.tzld.crawler.etl.model.dto.VideoInfoDto;
import com.tzld.crawler.etl.model.param.CrawlerVideoSendParam;
import com.tzld.crawler.etl.model.po.CrawlerVideo;
import com.tzld.crawler.etl.model.vo.CrawlerVideoVO;
import com.tzld.crawler.etl.model.vo.WxVideoVO;
import com.tzld.crawler.etl.service.EtlService;
import com.tzld.crawler.etl.service.SlsService;
import com.tzld.crawler.etl.service.feign.LongVideoFeign;
import com.tzld.crawler.etl.service.strategy.StrategyHandlerService;
import com.tzld.crawler.etl.util.FeishuUtils;
import com.tzld.crawler.etl.util.FileUtils;
import com.tzld.crawler.etl.util.MD5Util;
import com.tzld.crawler.etl.util.VideoUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;

/**
 * @author ehlxr
 * @since 2023-06-09 15:23.
 */
@Service
public class EtlServiceImpl implements EtlService {
    private final static Logger log = LoggerFactory.getLogger(EtlServiceImpl.class);
    private final StrategyHandlerService strategyHandlerService;
    private final AliyunOssManager aliyunOssManager;
    private final LongVideoFeign longVideoFeign;
    private final CrawlerVideoMapper crawlerVideoMapper;
    private final SlsService slsService;

    @Value("${download.file.path:/data/crawler/videos}")
    private String downloadPath;
    @Value("${aliyun.oss.video.bucket:art-pubbucket}")
    private String ossBucket;
    @Value("${env:}")
    private String env;
    @Value("${feishu.appid:}")
    private String feishuAppid;
    @Value("${feishu.appsecret:}")
    private String feishuAppsecret;
    @ApolloJsonValue("${feishu.sheet.token.map:{}}")
    private Map<String, String> feishuSheetTokenMap;
    @ApolloJsonValue("${feishu.sheet.id.map:{}}")
    private Map<String, String> feishuSheetIdMap;
    @ApolloJsonValue("${feishu.range.map:{}}")
    private Map<String, String> feishuRangeMap;
    @Value("${admin.cms.url:https://testadmin.piaoquantv.com/cms/post-detail/%d/info}")
    private String adminCmsUrl;
    @Value("${ffprobe.path:ffprobe}")
    private String ffprobePath;

    private Executor pool;

    public EtlServiceImpl(StrategyHandlerService strategyHandlerService, AliyunOssManager aliyunOssManager,
                          LongVideoFeign longVideoFeign, CrawlerVideoMapper crawlerVideoMapper, SlsService slsService) {
        this.strategyHandlerService = strategyHandlerService;
        this.aliyunOssManager = aliyunOssManager;
        this.longVideoFeign = longVideoFeign;
        this.crawlerVideoMapper = crawlerVideoMapper;
        this.slsService = slsService;
    }

    @Override
    public void deal(CrawlerVideoVO param) {
        String title = param.getVideoTitle();
        String platform = param.getPlatform();
        String strategy = param.getStrategy();
        try {
            // 1.策略应用
            // 获取用户选择策略
            List<String> stategies = Lists.newArrayList("titleScore");

            StrategyDataDto data = strategyHandlerService.execute(stategies, param);
            if (data == null) {
                return;
            }

            // 2.视频文件下载、上传 OSS、清理视频信息
            // 3.视频封面下载、上传 OSS、清理视频信息
            processVideo(data);

            // 4.视频发布
            CrawlerVideoSendParam request = new CrawlerVideoSendParam();
            request.setLoginUid(data.getUserId());
            request.setAppType(888888);
            request.setTitle(title);
            request.setVideoPath(data.getVideoOssPath());
            request.setCoverImgPath(data.getCoverOssPath());
            request.setTotalTime(data.getDuration());
            request.setVersionCode(1);
            request.setViewStatus(1);
            request.setCrawlerSrcId(data.getOutVideoId());
            request.setCrawlerSrcCode(platform.toUpperCase());
            LocalDateTime localDateTime = LocalDateTime.parse(data.getPublishTime(), DateTimeFormatter.ofPattern(Constant.STANDARD_FORMAT));
            request.setCrawlerSrcPublishTimestamp(localDateTime.toInstant(ZoneOffset.of("+8")).toEpochMilli());
            request.setCrawlerTaskTimestamp(System.currentTimeMillis());
            CommonResponse<WxVideoVO> response = longVideoFeign.crawlerVideoSend(request);
            log.info("crawler data send request: {}, response: {}", request, response);
            if (!response.isSuccess()) {
                throw new CommonException(ExceptionEnum.INVOKE_VIDEOAPI_ERROR, "invoke crawler data send failed!" + response);
            }

            // 5.视频元信息更新

            // 6.保存信息到数据库
            CrawlerVideo crawlerVideo = new CrawlerVideo();
            BeanUtils.copyProperties(data, crawlerVideo);
            crawlerVideo.setVideoId(response.getData().getId());
            String insertSql = "insert into crawler_video(video_id, user_id, out_user_id, platform, strategy, out_video_id, video_title," +
                    " cover_url, video_url, duration, publish_time, play_cnt, crawler_rule, width, height) values(" + crawlerVideo.getVideoId() + "," +
                    crawlerVideo.getUserId() + "," +
                    crawlerVideo.getOutUserId() + "," +
                    crawlerVideo.getPlatform() + "," +
                    crawlerVideo.getStrategy() + "," +
                    crawlerVideo.getOutVideoId() + "," +
                    crawlerVideo.getVideoTitle() + "," +
                    crawlerVideo.getCoverUrl() + "," +
                    crawlerVideo.getVideoUrl() + "," +
                    crawlerVideo.getDuration() + "," +
                    crawlerVideo.getPublishTime() + "," +
                    crawlerVideo.getPlayCnt() + "," +
                    crawlerVideo.getCrawlerRule() + "," +
                    crawlerVideo.getWidth() + "," +
                    crawlerVideo.getHeight() + ")";
            slsService.log("message", "insert_sql: " + insertSql, "crawler", platform, "mode", strategy);
            crawlerVideoMapper.insertSelective(crawlerVideo);
            slsService.log("message", "视频信息写入数据库成功", "crawler", platform, "mode", strategy);

            // 7.视频写入飞书
            pool.execute(() -> {
                try {
                    String sheetToken = feishuSheetTokenMap.get(platform);
                    if (Strings.isNullOrEmpty(sheetToken)) {
                        log.error("feishu.sheet.token.map config of {} is null", platform);
                        return;
                    }
                    String strategyType = Strings.nullToEmpty(param.getStrategyType());
                    String sheetId = feishuSheetIdMap.get(platform + strategyType);
                    if (Strings.isNullOrEmpty(sheetToken)) {
                        log.error("feishu.sheet.id.map config of {} is null", platform);
                        return;
                    }
                    String fsResp = FeishuUtils.insertRows(feishuAppid, feishuAppsecret, sheetToken,
                            sheetId, 1, 2);
                    log.debug("insert columns to feishu sheet response is {}", fsResp);

                    List<List<Object>> values = new ArrayList<>();
                    List<Object> value = new ArrayList<>();
                    DateTimeFormatter formatter = DateTimeFormatter.ofPattern(Constant.STANDARD_FORMAT);
                    value.add(data.getTitleScore());
                    value.add(crawlerVideo.getVideoId());
                    value.add(formatter.format(LocalDateTime.ofInstant(Instant.ofEpochMilli(System.currentTimeMillis()), ZoneId.systemDefault())));
                    value.add(strategy);
                    value.add(data.getOutVideoId());
                    value.add(data.getVideoTitle());
                    value.add(String.format(adminCmsUrl, crawlerVideo.getVideoId()));
                    value.add(data.getPlayCnt());
                    value.add(data.getCommentCnt());
                    value.add(data.getLikeCnt());
                    value.add(data.getShareCnt());
                    value.add(data.getDuration());
                    value.add(data.getWidth() + "*" + data.getHeight());
                    value.add(data.getPublishTime());
                    value.add(data.getUserName());
                    value.add(data.getOutUserId());
                    value.add(data.getAvatarUrl());
                    value.add(data.getCoverUrl());
                    value.add(data.getVideoUrl());
                    values.add(value);

                    fsResp = FeishuUtils.updateValues(feishuAppid, feishuAppsecret, sheetToken,
                            feishuRangeMap.get(platform + strategyType), values);
                    log.debug("update feishu sheet value response is {}", fsResp);

                    slsService.log("message", "视频已保存至云文档", "crawler", platform, "mode", strategy);
                } catch (Exception e) {
                    log.error("save data to feishu sheet error. platform {}, strategy {}", platform, strategy, e);
                }
            });
        } catch (Exception e) {
            log.error("etl server deal {} failed.", param, e);
            throw new CommonException(ExceptionEnum.SYSTEM_ERROR, "etl server deal error: " + e.getMessage());
        }
    }

    private void processVideo(StrategyDataDto data) throws IOException {
        String title = data.getVideoTitle();
        String platform = data.getPlatform();
        String strategy = data.getStrategy();

        // 视频文件下载、上传 OSS
        String videoPath = url2oss(data.getVideoUrl(), "longvideo/crawler_local/data", title, platform, strategy);
        data.setVideoOssPath(videoPath);
        String tempFilePath = downloadPath + File.separator + videoPath;
        if (data.getDuration() == null || data.getDuration() <= 0
                || data.getWidth() == null || data.getWidth() <= 0
                || data.getHeight() == null || data.getHeight() <= 0) {
            VideoInfoDto videoInfo = VideoUtils.getVideoInfo(tempFilePath, ffprobePath);
            if (videoInfo != null) {
                data.setDuration(videoInfo.getDuration());
                data.setWidth(videoInfo.getWidth());
                data.setHeight(videoInfo.getHeight());
            }
        }

        // 文件清理
        log.info("begin delete: {}", tempFilePath);
        Files.deleteIfExists(Paths.get(new File(tempFilePath).getPath()));

        // 视频封面下载、上传 OSS
        String coverPath = url2oss(data.getCoverUrl(), "longvideo/crawler_local/image", title, platform, strategy);
        data.setCoverOssPath(coverPath);
        tempFilePath = downloadPath + File.separator + videoPath;

        // 文件清理
        log.info("begin delete: {}", tempFilePath);
        Files.deleteIfExists(Paths.get(new File(tempFilePath).getPath()));
    }

    private String url2oss(String fileUrl, String filePath, String title, String crawler, String mode) throws IOException {
        String titleMmd5 = MD5Util.md5(title);
        String curDate = LocalDate.now().format(DateTimeFormatter.BASIC_ISO_DATE);
        String videoFilePath = filePath + File.separator + env + File.separator + curDate;
        String localFilePath = downloadPath + File.separator + videoFilePath;

        File localFile = new File(localFilePath);
        if (!localFile.exists()) {
            if (!localFile.mkdirs()) {
                // throw new CommonException(ExceptionEnum.SYSTEM_ERROR, "mkdir " + localFilePath + " error!");
                log.warn("mkdir {} failed", localFilePath);
            }
        }

        // 下载文件
        int retry = 0;
        boolean downloadFlag = false;
        while (retry < 3) {
            retry++;
            downloadFlag = FileUtils.download(fileUrl, localFilePath + File.separator + titleMmd5);
            if (downloadFlag) {
                break;
            }
            log.warn("download file failed. retry times: {}, from file {} to {}. ", fileUrl, filePath, retry);
        }
        if (!downloadFlag) {
            throw new CommonException(ExceptionEnum.SYSTEM_ERROR, "download file " + fileUrl + " failed!");
        }

        // 文件上传 OSS
        slsService.log("message", "开始上传视频... ", "crawler", crawler, "mode", mode);
        log.info("begin upload: {}/{} to oss key: {}/{}", localFilePath, titleMmd5, videoFilePath, titleMmd5);
        aliyunOssManager.putObject(ossBucket, videoFilePath + File.separator + titleMmd5, Files.newInputStream(Paths.get(localFilePath + File.separator + titleMmd5)));

        return videoFilePath + File.separator + titleMmd5;
    }

    @PostConstruct
    public void init() {
        pool = Executors.newFixedThreadPool(1);
    }
}
